Lei Chen raised an issue with CLOCK_MONOTONIC_COARSE seeing
time inconsistencies.

Lei tracked down that this was being caused by the adjustment
  tk->tkr_mono.xtime_nsec -= offset;

which is made to compensate for the unaccumulated cycles in
offset when the mult value is adjusted forward, so that
the non-_COARSE clockids don't see inconsistencies.

However, the _COARSE clockids don't use the mult*offset value
in their calculations, so this subtraction can cause the
_COARSE clock ids to jump back a bit.

Now, by design, this negative adjustment should be fine, because
the logic run from timekeeping_adjust() is done after we
accumulate approx mult*interval_cycles into xtime_nsec.
The accumulated (mult*interval_cycles) will be larger then the
(mult_adj*offset) value subtracted from xtime_nsec, and both
operations are done together under the tk_core.lock, so the net
change to xtime_nsec should always be positive.

However, do_adjtimex() calls into timekeeping_advance() as well,
since we want to apply the ntp freq adjustment immediately.
In this case, we don't return early when the offset is smaller
then interval_cycles, so we don't end up accumulating any time
into xtime_nsec. But we do go on to call timekeeping_adjust(),
which modifies the mult value, and subtracts from xtime_nsec
to correct for the new mult value.

Here because we did not accumulate anything, we have a window
where the _COARSE clockids that don't utilize the mult*offset
value, can see an inconsistency.

So to fix this, rework the timekeeping_advance() logic a bit
so that when we are called from do_adjtimex() and the offset
is smaller then cycle_interval, that we call
timekeeping_forward(), to first accumulate the sub-interval
time into xtime_nsec. Then with no unaccumulated cycles in
offset, we can do the mult adjustment without worry of the
subtraction having an impact.

NOTE: This was implemented as a potential alternative to
Thomas' approach here:
   https://lore.kernel.org/lkml/87cyej5rid.ffs@tglx/

And similarly, it needs some additional review and testing,
as it was developed while packing for conference travel.

Cc: Thomas Gleixner <t...@linutronix.de>
Cc: Stephen Boyd <sb...@kernel.org>
Cc: Anna-Maria Behnsen <anna-ma...@linutronix.de>
Cc: Frederic Weisbecker <frede...@kernel.org>
Cc: Shuah Khan <sh...@kernel.org>
Cc: Miroslav Lichvar <mlich...@redhat.com>
Cc: linux-kselft...@vger.kernel.org
Cc: kernel-t...@android.com
Cc: Lei Chen <lei.c...@smartx.com>
Fixes: da15cfdae033 ("time: Introduce CLOCK_REALTIME_COARSE")
Reported-by: Lei Chen <lei.c...@smartx.com>
Closes: 
https://lore.kernel.org/lkml/20250310030004.3705801-1-lei.c...@smartx.com/
Diagnosed-by: Thomas Gleixner <t...@linutronix.de>
Signed-off-by: John Stultz <jstu...@google.com>
---
 kernel/time/timekeeping.c | 87 ++++++++++++++++++++++++++++-----------
 1 file changed, 62 insertions(+), 25 deletions(-)

diff --git a/kernel/time/timekeeping.c b/kernel/time/timekeeping.c
index 1e67d076f1955..6f3a145e7b113 100644
--- a/kernel/time/timekeeping.c
+++ b/kernel/time/timekeeping.c
@@ -682,18 +682,18 @@ static void timekeeping_update_from_shadow(struct tk_data 
*tkd, unsigned int act
 }
 
 /**
  * timekeeping_forward_now - update clock to the current time
  * @tk:                Pointer to the timekeeper to update
+ * @cycle_now:  Current clocksource read value
  *
  * Forward the current clock to update its state since the last call to
  * update_wall_time(). This is useful before significant clock changes,
  * as it avoids having to deal with this time offset explicitly.
  */
-static void timekeeping_forward_now(struct timekeeper *tk)
+static void timekeeping_forward(struct timekeeper *tk, u64 cycle_now)
 {
-       u64 cycle_now, delta;
+       u64 delta;
 
-       cycle_now = tk_clock_read(&tk->tkr_mono);
        delta = clocksource_delta(cycle_now, tk->tkr_mono.cycle_last, 
tk->tkr_mono.mask,
                                  tk->tkr_mono.clock->max_raw_delta);
        tk->tkr_mono.cycle_last = cycle_now;
@@ -710,6 +710,21 @@ static void timekeeping_forward_now(struct timekeeper *tk)
        }
 }
 
+/**
+ * timekeeping_forward_now - update clock to the current time
+ * @tk:                Pointer to the timekeeper to update
+ *
+ * Forward the current clock to update its state since the last call to
+ * update_wall_time(). This is useful before significant clock changes,
+ * as it avoids having to deal with this time offset explicitly.
+ */
+static void timekeeping_forward_now(struct timekeeper *tk)
+{
+       u64 cycle_now = tk_clock_read(&tk->tkr_mono);
+
+       timekeeping_forward(tk, cycle_now);
+}
+
 /**
  * ktime_get_real_ts64 - Returns the time of day in a timespec64.
  * @ts:                pointer to the timespec to be set
@@ -2151,6 +2166,45 @@ static u64 logarithmic_accumulation(struct timekeeper 
*tk, u64 offset,
        return offset;
 }
 
+static u64 timekeeping_accumulate(struct timekeeper *tk, u64 now, u64 offset,
+                                 unsigned int *clock_set)
+{
+       struct timekeeper *real_tk = &tk_core.timekeeper;
+       int shift = 0, maxshift;
+
+       /*
+        * If we have a sub-cycle_interval offset, we
+        * are likely doing a TK_FREQ_ADJ, so accumulate
+        * everything so we don't have a remainder offset
+        * when later adjusting the multiplier
+        */
+       if (offset < real_tk->cycle_interval) {
+               timekeeping_forward(tk, now);
+               *clock_set = 1;
+               return 0;
+       }
+
+       /*
+        * With NO_HZ we may have to accumulate many cycle_intervals
+        * (think "ticks") worth of time at once. To do this efficiently,
+        * we calculate the largest doubling multiple of cycle_intervals
+        * that is smaller than the offset.  We then accumulate that
+        * chunk in one go, and then try to consume the next smaller
+        * doubled multiple.
+        */
+       shift = ilog2(offset) - ilog2(tk->cycle_interval);
+       shift = max(0, shift);
+       /* Bound shift to one less than what overflows tick_length */
+       maxshift = (64 - (ilog2(ntp_tick_length()) + 1)) - 1;
+       shift = min(shift, maxshift);
+       while (offset >= tk->cycle_interval) {
+               offset = logarithmic_accumulation(tk, offset, shift, clock_set);
+               if (offset < tk->cycle_interval << shift)
+                       shift--;
+       }
+       return offset;
+}
+
 /*
  * timekeeping_advance - Updates the timekeeper to the current time and
  * current NTP tick length
@@ -2160,8 +2214,7 @@ static bool timekeeping_advance(enum timekeeping_adv_mode 
mode)
        struct timekeeper *tk = &tk_core.shadow_timekeeper;
        struct timekeeper *real_tk = &tk_core.timekeeper;
        unsigned int clock_set = 0;
-       int shift = 0, maxshift;
-       u64 offset;
+       u64 cycle_now, offset;
 
        guard(raw_spinlock_irqsave)(&tk_core.lock);
 
@@ -2169,7 +2222,8 @@ static bool timekeeping_advance(enum timekeeping_adv_mode 
mode)
        if (unlikely(timekeeping_suspended))
                return false;
 
-       offset = clocksource_delta(tk_clock_read(&tk->tkr_mono),
+       cycle_now = tk_clock_read(&tk->tkr_mono);
+       offset = clocksource_delta(cycle_now,
                                   tk->tkr_mono.cycle_last, tk->tkr_mono.mask,
                                   tk->tkr_mono.clock->max_raw_delta);
 
@@ -2177,24 +2231,7 @@ static bool timekeeping_advance(enum 
timekeeping_adv_mode mode)
        if (offset < real_tk->cycle_interval && mode == TK_ADV_TICK)
                return false;
 
-       /*
-        * With NO_HZ we may have to accumulate many cycle_intervals
-        * (think "ticks") worth of time at once. To do this efficiently,
-        * we calculate the largest doubling multiple of cycle_intervals
-        * that is smaller than the offset.  We then accumulate that
-        * chunk in one go, and then try to consume the next smaller
-        * doubled multiple.
-        */
-       shift = ilog2(offset) - ilog2(tk->cycle_interval);
-       shift = max(0, shift);
-       /* Bound shift to one less than what overflows tick_length */
-       maxshift = (64 - (ilog2(ntp_tick_length())+1)) - 1;
-       shift = min(shift, maxshift);
-       while (offset >= tk->cycle_interval) {
-               offset = logarithmic_accumulation(tk, offset, shift, 
&clock_set);
-               if (offset < tk->cycle_interval<<shift)
-                       shift--;
-       }
+       offset = timekeeping_accumulate(tk, cycle_now, offset, &clock_set);
 
        /* Adjust the multiplier to correct NTP error */
        timekeeping_adjust(tk, offset);
@@ -682,18 +682,18 @@ static void timekeeping_update_from_shadow(struct tk_data 
 
 /**
- * timekeeping_forward_now - update clock to the current time
+ * timekeeping_forward - update clock to given cycle now value
  * @tk:                Pointer to the timekeeper to update
-- 
2.49.0.rc1.451.g8f38331e32-goog


Reply via email to